/*-------------------------------------------------------------
 *  Copyright (c) 2011-2017, MSS ComBlock. All rights reserved.
 *
 * 	Filename: RF.c
 *  Last modified: 2/18/17
 *  Authors:
 *  	Alain Zarembowitch / MSS ComBlock
 *  Version: 6
 *  Inheritance: N/A
 *  Description: RF components driver
 *
 *  Rev 1.1 6/1/11 AZ
 *  Moved RF gain control 2 from PWM 1.2 to PWM 1.4 (reason: hardware version change from prototype to production)
 *
 *  Rev 2 4/25/12 AZ
 *  Corrected bug in program_tuner_pll
 *
 *  Rev 3 3/9/13 AZ
 *  Read mixer power detector, even when AGC loop is open
 *  reduced AGC threshold to prevent saturation (wideband noise + signal)
 *  Note: make sure shield is soldered on all 4 edges to prevent oscillations.
 *
 *  Rev 4 4/15/14 AZ
 *  Rewrote AD7298_rw to shave a few us (3 us).
 *  Added ISRs triggered by falling and rising edges of RX_TXN to speed up the switching
 *
 *  Rev 5 8/9/14 AZ
 *  Reduced the rx RF synthesizer level from -1 to -4 dBm to minimize leakage (will be fixed in
 *  next hardware rev)
 *  Changed the RF AGC 'too hot' condition: RF gain is reduced when power detected
 *  at the mixer exceeds the threshold (normal) AND IF power detected too small.
 *  This is to prevent AGC deadlock when LO leakage is in-band and exceeds rf signal level.
 *
 *  Rev 6 2/18/17 AZ
 *  Added 4009 NASA special configuration (divided 1:8 LO)
 *  Phase noise is SIGNIFICANTLY better on the COM-4009 board when RF frequency doubling and :2 are disabled.
 -------------------------------------------------------------*/
#include "main.h"
#include "rf.h"
#include "utils.h"

/*-------- change this section according to the schematics ----------------------------------------*/
inline long synth_check_lock() {
	return FIO1PIN & 0x00040000; // pin P1[18] ADF4351 synthesizer lock status
}


/*-------- end of schematics-dependent code ----------------------------------------*/



unsigned long get_freq(long index) {
	// reconstruct the RF frequency synthesizer frequency from the control registers.
	// index range 0 - 7
	// Returns the frequency expressed in Hz.
	long reg_no;
	unsigned long freq;
	long i;

	// points to control register with MSB
	if (index == 0)
		reg_no = 3; // frequency in REG3-0
	else if (index < 8)
		reg_no = 6 + 4 * index;
	else
		return 0; // invalid index

	for (i = 0; i < 4; i++) {
		freq <<= 8;
		freq &= 0xFFFFFF00;
		freq |= ((unsigned long) regv[reg_no--]) & 0x000000FF;
	}
	return freq;
}

void adf4351_reg(long regval, int synth_no) {
	// write 32 serial bits to the ADF4351 Integrated synthesizer and VCO

	int i;
	if(synth_no == 2){
		PIN_SYNTH2_CLK_L(); // always start transaction with clock low
		PIN_SYNTH2_LE_L(); // Latch enable - low  (initialized high at power up)

		for (i = 0; i < 32; i++) {
			if (regval & 0x80000000) {
				PIN_SYNTH2_DATA_H(); //data high
				PIN_SYNTH2_CLK_H(); //clock high
			} else {
				PIN_SYNTH2_DATA_L(); //data low
				PIN_SYNTH2_CLK_H(); //clock high
			}

			regval <<= 1;
			PIN_SYNTH2_CLK_L(); //clock low
		}

		PIN_SYNTH2_LE_H(); // Latch disable - high

	}
}

//-----------------------------------------------------------------------
//-- RF synthesizer profiles
//
// init_tuner_pll(,1);  // 26 MHz frequency reference
// init_tuner_pll(,2);  // 19.2 MHz frequency reference, integer-N
// init_tuner_pll(,3);  // 19.2 MHz frequency reference, fractional-N (modified from COM-3011 with 20 MHz frequency reference)
// init_tuner_pll(,4);  // 10 MHz frequency reference, fractional-N
// init_tuner_pll(,5);  // 10 MHz frequency reference, integer-N, Pfd = 0.5MHz


void init_tuner_pll(int synth_no, char profile_no) {
	// one-time ADF4351 initialization.
	// program registers in sequence 5,4,3,2,1,0
	unsigned long regval;

	// register R5
	regval = 0x00580005;
	// LD pin = digital lock detect
	adf4351_reg(regval, synth_no);

	if(profile_no == 1){
		// 26 MHz frequency reference. Pfd = 13 MHz. Integer-N

		// register R3
		regval = 0x00E48283;
		// clock divider value = 50
		// fast lock enable
		// cycle slip reduction enable
		// charge cancellation
		// band select clock mode high
		adf4351_reg(regval, synth_no);

		// register R2
	//	regval = 0x1806AFC2;   // R counter = 26, reference doubler disabled, reference divide by 2 disabled
		regval = 0x19036FC2;   // R counter = 13, reference doubler disabled, reference divide by 2 enabled


		// counter reset disabled
		// cp three-state disabled
		// not powered down
		// pd positive polarity, 10ns LDP
		// int-N, icp = 2.5mA
		// double buffer enabled
		// muxout = digital lock detect
		// low phase noise mode
		adf4351_reg(regval, synth_no);

		// register R1
		regval = 0x00008011; // 4/5 prescaler,
		adf4351_reg(regval, synth_no);
	} else if (profile_no == 2){
		// 19.2 MHz frequency reference. Pfd = 400 kHz. Integer-N

		// register R3
		regval = 0x00E08023;
		// clock divider value = 4
		// fast lock enable
		// cycle slip reduction disable
		// charge cancellation
		// band select clock mode high
		adf4351_reg(regval, synth_no);

		// register R2
		regval = 0x1B0C21C2;   // R counter = 48, reference doubler enabled, reference divide by 2 enabled
		// counter reset disabled
		// cp three-state disabled
		// not powered down
		// pd positive polarity, 6ns LDP
		// int-N, icp = 0.31mA
		// double buffer enabled
		// muxout = digital lock detect
		// low phase noise mode
		adf4351_reg(regval, synth_no);

		// register R1
		regval = 0x08008011; // 8/9 prescaler,
		adf4351_reg(regval, synth_no);

	} else if (profile_no == 3){
		// 19.2 MHz frequency reference. Pfd = 5 kHz. Fractional-N

		// register R3
		regval = 0x00008013;
		// clock divider value = 400uS * 5 KHz = 2
		// fast clock enable
		// cycle slip reduction disabled (seems to increase phase noise a lot! when enabled)
		adf4351_reg(regval, synth_no);

		// register R2
		// double 19.2 MHz ref clock for better phase noise)
		regval = 0x18006E42; //  preference: low noise mode
		// counter reset disabled
		// cp three-state disabled
		// not powered down
		// pd positive polarity, 10ns LDP
		// frac-N, icp = 2.5ma
		// double buffer enabled
		// 10-bit R counter = 1, NO ref doubling NO 1:2
		// muxout = digital lock detect
		// low phase noise mode
		adf4351_reg(regval, synth_no);

		// register R1
		regval = 0x0800F801; // 8/9 prescaler, MOD = 3840
		adf4351_reg(regval, synth_no);
	} else if (profile_no == 4){
		// 10 MHz frequency reference. Pfd = 5 kHz. Fractional-N

		// register R3
		regval = 0x00008013;
		// clock divider value = 400uS * 5 KHz = 2
		// fast clock enable
		// cycle slip reduction disabled (seems to increase phase noise a lot! when enabled)
		adf4351_reg(regval, synth_no);

		// register R2
		// double 10 MHz ref clock for better phase noise)
		regval = 0x18006E42; //  preference: low noise mode
		// counter reset disabled
		// cp three-state disabled
		// not powered down
		// pd positive polarity, 10ns LDP
		// frac-N, icp = 2.5ma
		// double buffer enabled
		// 10-bit R counter = 1, NO ref doubling NO 1:2
		// muxout = digital lock detect
		// low phase noise mode
		adf4351_reg(regval, synth_no);

		// register R1
		regval = 0x0800BE81; // 8/9 prescaler, MOD = 2000
		adf4351_reg(regval, synth_no);

	} else if (profile_no == 5){
		// 10 MHz frequency reference. Pfd = 500 KHz. Integer-N
		// Note: fractional-N may have better phase noise (based on 4009)

		// register R3
		regval = 0x00E08643;
		// clock divider value = 400uS * 0.5 MHz = 200
		// fast clock enable, select high band, anti-backlash 3ns, charge cancellation
		// cycle slip reduction disabled (seems to increase phase noise a lot! when enabled)
		adf4351_reg(regval, synth_no);

		// register R2
		// double 10 MHz ref clock for better phase noise)
		regval = 0x1902afc2; //  preference: low noise mode
		// counter reset disabled
		// cp three-state disabled
		// not powered down
		// pd positive polarity, 10ns LDP
		// int-N, icp = 2.5ma
		// double buffer enabled
		// 10-bit R counter = 10, no doubling followed by 1:2
		// muxout = digital lock detect
		// low phase noise mode
		adf4351_reg(regval, synth_no);

		// register R1
		regval = 0x00008011; // 4/5 prescaler,
		adf4351_reg(regval, synth_no);

	}
	return;
}

void program_tuner_pll(int synth_no, char profile_no) {
	// program RF synthesizer with default frequency
	// One must call init_tuner_pll() once prior to using this function (one-time).

	unsigned long regval, f, fvcodiv2, fi;
	unsigned short nint, nfrac;

	// construct frequency word (in Hz) from control registers
	if(synth_no == 2){
		// baseband to transmit frequency
		// Tx frequency index is in REG6[7:5]
		fi = (regv[6] >> 5) & 0x07;
		f = get_freq(fi); // frequency in Hz
	}



	if(profile_no == 1){
		// 26 MHz frequency reference. Pfd = 13 MHz. Integer-N

		// program registers in sequence 5,4,3,2,1,0
		// register R4
		// convert tuner_freq into parameters understandable by the ADF4351
		// multiply turn tuner_freq to get a VCO frequency before the final output divider
		// Disable auxiliary output (RF_OUTB unused)
		// Set output LO power to -1dBm (trying to minimize LO leakage to the RF input).
		if (f >= 2200000000) {
			regval = 0x0080842C; // fout >= 2.2 GHz
			fvcodiv2 = f >> 1; // VCO frequency/2 (to prevent overflow later in the computation)
		} else if (f >= 1100000000) {
			regval = 0x0090842C; // fout >= 1.1 GHz
			fvcodiv2 = f;
		} else if (f >= 550000000) {
			regval = 0x00A0842C; // fout >= 550 MHz
			fvcodiv2 = f << 1;
		} else if (f >= 275000000) {
			regval = 0x00B0842C; // fout >= 275 MHz
			fvcodiv2 = f << 2;
		} else if (f >= 137500000) {
			regval = 0x00C0842C; // fout >= 137.5 MHz
			fvcodiv2 = f << 3;
		} else if (f >= 68750000) {
			regval = 0x00D0842C; // fout >= 68.75 MHz
			fvcodiv2 = f << 4;
		} else {
			regval = 0x00E0842C; // fout >= 34.375 MHz
			fvcodiv2 = f << 5;
		}
		// feedback select = fundamental
		// band select clock divider = 26MHz/125KHz max = 208. use 220 (xDC).
		// vco powered up, mute till lock detect,
		// aux output disabled, divided outputs
		adf4351_reg(regval, synth_no);

		// register R0
		// reference frequency is 1 MHz
		nint = (unsigned short) (fvcodiv2 / (1000000/2));
		// integer-N PLL
		regval = (unsigned long) (nint);
		regval <<= 15;
		regval = regval & 0x7FFF8000;
		adf4351_reg(regval, synth_no);
	} else if (profile_no == 2){
		// 19.2 MHz frequency reference. Pfd = 200 kHz. Integer-N

		// program registers in sequence 5,4,3,2,1,0
		// register R4
		// convert tuner_freq into parameters understandable by the ADF4351
		// multiply turn tuner_freq to get a VCO frequency before the final output divider
		// Disable auxiliary output (RF_OUTB unused)
		// Set output LO power to -1dBm (trying to minimize LO leakage to the RF input).
		if (f >= 2200000000) {
			regval = 0x0080842C; // fout >= 2.2 GHz
			fvcodiv2 = f >> 1; // VCO frequency/2 (to prevent overflow later in the computation)
		} else if (f >= 1100000000) {
			regval = 0x0090842C; // fout >= 1.1 GHz
			fvcodiv2 = f;
		} else if (f >= 550000000) {
			regval = 0x00A0842C; // fout >= 550 MHz
			fvcodiv2 = f << 1;
		} else if (f >= 275000000) {
			regval = 0x00B0842C; // fout >= 275 MHz
			fvcodiv2 = f << 2;
		} else if (f >= 137500000) {
			regval = 0x00C0842C; // fout >= 137.5 MHz
			fvcodiv2 = f << 3;
		} else if (f >= 68750000) {
			regval = 0x00D0842C; // fout >= 68.75 MHz
			fvcodiv2 = f << 4;
		} else {
			regval = 0x00E0842C; // fout >= 34.375 MHz
			fvcodiv2 = f << 5;
		}
		// feedback select = fundamental
		// band select clock divider = 26MHz/125KHz max = 208. use 220 (xDC).
		// vco powered up, mute till lock detect,
		// aux output disabled, divided outputs
		adf4351_reg(regval, synth_no);

		// register R0
		// reference frequency is 400 kHz
		nint = (unsigned short) (fvcodiv2 / (400000/2));
		// integer-N PLL
		regval = (unsigned long) (nint);
		regval <<= 15;
		regval = regval & 0x7FFF8000;
		adf4351_reg(regval, synth_no);

	} else if (profile_no == 3){
		// 19.2 MHz frequency reference. Pfd = 5 kHz. Fractional-N

		// register R4
		// convert tuner_freq into parameters understandable by the ADF4351
		// multiply turn tuner_freq to get a VCO frequency before the final output divider
		// Disable auxiliary output (RF_OUTB unused)
		// Set output LO power to -1dBm (trying to minimize LO leakage to the RF input).
		if (f >= 2200000000) {
			if(synth_no == 1) {
				// Rx. Reduce LO level to -4 dBm(see DR217)
				regval = 0x008C8424; // fout >= 2.2 GHz
			} else {
				regval = 0x008C842C; // fout >= 2.2 GHz
			}
			fvcodiv2 = f >> 1; // VCO frequency/2 (to prevent overflow later in the computation)
		} else if (f >= 1100000000) {
			if(synth_no == 1) {
				// Rx. Reduce LO level to -4 dBm(see DR217)
				regval = 0x009C8424; // fout >= 2.2 GHz
			} else {
				regval = 0x009C842C; // fout >= 1.1 GHz
			}
			fvcodiv2 = f;
		} else if (f >= 550000000) {
			regval = 0x00AC842C; // fout >= 550 MHz
			fvcodiv2 = f << 1;
		} else if (f >= 275000000) {
			regval = 0x00BC842C; // fout >= 275 MHz
			fvcodiv2 = f << 2;
		} else if (f >= 137500000) {
			regval = 0x00CC842C; // fout >= 137.5 MHz
			fvcodiv2 = f << 3;
		} else if (f >= 68750000) {
			regval = 0x00DC842C; // fout >= 68.75 MHz
			fvcodiv2 = f << 4;
		} else {
			regval = 0x00EC842C; // fout >= 34.375 MHz
			fvcodiv2 = f << 5;
		}
		// feedback select = fundamental
		// band select clock divider = 19.2MHz/125KHz max = 154. use 200 (xC8).
		// vco powered up, mute till lock detect,
		// aux output disabled, divided outputs
		adf4351_reg(regval, synth_no);

		// register R0
		// reference frequency is 19.2 MHz
		nint = (unsigned short) (fvcodiv2 / 9600000);
		fvcodiv2 -= nint * 9600000;
		// frac step size is 5KHz
		nfrac = (unsigned short) (fvcodiv2 * 2 / 5000);
		regval = (unsigned long) (nint);
		regval <<= 12;
		regval = regval & 0x0FFFF000;
		regval |= ((unsigned long) (nfrac)) & 0x000FFF;
		regval <<= 3;
		regval &= 0x7FFFFFF8;
		adf4351_reg(regval, synth_no);

	} else if (profile_no == 4){
		// 10 MHz frequency reference. Pfd = 5 kHz. Fractional-N

		// register R4
		// convert tuner_freq into parameters understandable by the ADF4351
		// multiply turn tuner_freq to get a VCO frequency before the final output divider
		// Disable auxiliary output (RF_OUTB unused)
		// Set output LO power to -1dBm (trying to minimize LO leakage to the RF input).
		if (f >= 2200000000) {
			regval = 0x008C842C; // fout >= 2.2 GHz
			fvcodiv2 = f >> 1; // VCO frequency/2 (to prevent overflow later in the computation)
		} else if (f >= 1100000000) {
			regval = 0x009C842C; // fout >= 1.1 GHz
			fvcodiv2 = f;
		} else if (f >= 550000000) {
			regval = 0x00AC842C; // fout >= 550 MHz
			fvcodiv2 = f << 1;
		} else if (f >= 275000000) {
			regval = 0x00BC842C; // fout >= 275 MHz
			fvcodiv2 = f << 2;
		} else if (f >= 137500000) {
			regval = 0x00CC842C; // fout >= 137.5 MHz
			fvcodiv2 = f << 3;
		} else if (f >= 68750000) {
			regval = 0x00DC842C; // fout >= 68.75 MHz
			fvcodiv2 = f << 4;
		} else {
			regval = 0x00EC842C; // fout >= 34.375 MHz
			fvcodiv2 = f << 5;
		}
		// feedback select = fundamental
		// band select clock divider = 10MHz/125KHz max = 80. use 100.
		// vco powered up, mute till lock detect,
		// aux output disabled, divided outputs
		adf4351_reg(regval, synth_no);

		// register R0
		// reference frequency is 10 MHz
		nint = (unsigned short) (fvcodiv2 / 5000000);
		fvcodiv2 -= nint * 5000000;
		// frac step size is 5KHz
		nfrac = (unsigned short) (fvcodiv2 * 2 / 5000);
		regval = (unsigned long) (nint);
		regval <<= 12;
		regval = regval & 0x0FFFF000;
		regval |= ((unsigned long) (nfrac)) & 0x000FFF;
		regval <<= 3;
		regval &= 0x7FFFFFF8;
		adf4351_reg(regval, synth_no);

	} else if (profile_no == 5){
		// 19.2 MHz frequency reference. Pfd = 5 kHz. Fractional-N
		// RF_OUTB used. 2.2 - 4.4 GHz only
		// RF_OUTA divided by 8

		// register R4
		// convert tuner_freq into parameters understandable by the ADF4351
		// multiply turn tuner_freq to get a VCO frequency before the final output divider
		// Disable auxiliary output (RF_OUTB unused)
		// Set output LO power to -1dBm (trying to minimize LO leakage to the RF input).
		if (f >= 2200000000) {
			regval = 0x00BC877C; // fout >= 2.2 GHz, aux output, main divided by 8
			fvcodiv2 = f >> 1; // VCO frequency/2 (to prevent overflow later in the computation)
		}
		// feedback select = fundamental
		// band select clock divider = 19.2MHz/125KHz max = 154. use 200 (xC8).
		// vco powered up, mute till lock detect,
		// aux output enabled, main output enabled divided by 8
		adf4351_reg(regval, synth_no);

		// register R0
		// reference frequency is 19.2 MHz
		nint = (unsigned short) (fvcodiv2 / 9600000);
		fvcodiv2 -= nint * 9600000;
		// frac step size is 5KHz
		nfrac = (unsigned short) (fvcodiv2 * 2 / 5000);
		regval = (unsigned long) (nint);
		regval <<= 12;
		regval = regval & 0x0FFFF000;
		regval |= ((unsigned long) (nfrac)) & 0x000FFF;
		regval <<= 3;
		regval &= 0x7FFFFFF8;
		adf4351_reg(regval, synth_no);

	} else if (profile_no == 6){
		// 10 MHz frequency reference. Pfd = 5 kHz. Fractional-N
		// RF_OUTB used. 2.2 - 4.4 GHz only
		// RF_OUTA divided by 8

		// register R4
		// convert tuner_freq into parameters understandable by the ADF4351
		// multiply turn tuner_freq to get a VCO frequency before the final output divider
		// Set output LO power to -1dBm (trying to minimize LO leakage to the RF input).
		if (f >= 2200000000) {
			regval = 0x00BC877C; // fout >= 2.2 GHz, aux output, main divided by 8
			fvcodiv2 = f >> 1; // VCO frequency/2 (to prevent overflow later in the computation)
		}
		// feedback select = fundamental
		// band select clock divider = 10MHz/125KHz max = 80. use 100.
		// vco powered up, mute till lock detect,
		// aux output disabled, divided outputs
		adf4351_reg(regval, synth_no);

		// register R0
		// reference frequency is 10 MHz
		nint = (unsigned short) (fvcodiv2 / 5000000);
		fvcodiv2 -= nint * 5000000;
		// frac step size is 5KHz
		nfrac = (unsigned short) (fvcodiv2 * 2 / 5000);
		regval = (unsigned long) (nint);
		regval <<= 12;
		regval = regval & 0x0FFFF000;
		regval |= ((unsigned long) (nfrac)) & 0x000FFF;
		regval <<= 3;
		regval &= 0x7FFFFFF8;
		adf4351_reg(regval, synth_no);

	} else if (profile_no == 7){
		// 10 MHz frequency reference. Pfd = 500 KHz. Integer-N
		// RF_OUTB used. 2.2 - 4.4 GHz only
		// RF_OUTA divided by 8

		// register R4
		// convert tuner_freq into parameters understandable by the ADF4351
		// multiply turn tuner_freq to get a VCO frequency before the final output divider
		// Set rf LO power to -1dBm (trying to minimize LO leakage to the RF input).
		// Set aux output power to +5dBm (lower frequency, need max output).
		if (f >= 2200000000) {
			regval = 0x00B047eC; // fout >= 2.2 GHz, aux output, main divided by 8
			fvcodiv2 = f >> 1; // VCO frequency/2 (to prevent overflow later in the computation)
		}
		// feedback select = fundamental
		// band select clock divider = 10MHz/125KHz max = 80. use 100.
		// vco powered up, mute till lock detect,
		// aux output disabled, divided outputs
		adf4351_reg(regval, synth_no);

		// register R0
		// reference frequency is 0.5 MHz
		nint = (unsigned short) (fvcodiv2 / (500000/2));
		// integer-N PLL
		regval = (unsigned long) (nint);
		regval <<= 15;
		regval = regval & 0x7FFF8000;
		adf4351_reg(regval, synth_no);

	}
}

inline void TX_EN_FREQ_HOP_R_ISR(){
	// Interrupt service routine when PDN or TX_EN interface signal rising edge

	// TX_EN rising edge
	if((IO0_INT_STAT_R & 0x00008000) != 0){
		// PDN rising edge
		// TODO in rev1
	}
	if((IO0_INT_STAT_R & 0x00010000) != 0){
		// TX_EN rising edge
		PIN_RFOUT_ENB_L();	// enable the RF modulator output (active low)
		PIN_SYNTH2_ENABLE_H();	// enable the RF synthesizer output
	}
}
inline void TX_EN_FREQ_HOP_F_ISR(){
	int freq_sel;
	// Interrupt service routine when PDN, TX_EN or FREQ_HOP interface signals falling edge

	// PDN, TX_EN or FREQ_HOP?
	if((IO0_INT_STAT_F & 0x00008000) != 0){
		// PDN falling edge
		// TODO in rev1
	}
	if((IO0_INT_STAT_F & 0x00010000) != 0){
		// TX_EN falling edge
		PIN_RFOUT_ENB_H();	// disable the RF modulator output
		PIN_SYNTH2_ENABLE_L();	// disable the RF synthesizer output
	}
	if((IO0_INT_STAT_F & 0x00020000) != 0){
		// FREQ_HOP falling edge
		freq_sel = (regv[6] >> 5) & 0x07;
		// increment the frequency selection
		freq_sel++;
		if(freq_sel >= (regv[35] & 0x0F)){
			freq_sel = 0;
		}
		// store value back in regv[6]
		regv[6] &= ~0x00E0;
		regv[6] |= (freq_sel << 5) & 0x00E0;
		// RF frequency synthesizer (ADF4351) initialization
		if(VCTCXO_REF) {
			// 19.2 MHz internal (TCXO) frequency reference
			init_tuner_pll(2,3); // one time initialization
			program_tuner_pll(2,3); // Program the default frequency
		} else {
			// 10 MHz external frequency reference
			init_tuner_pll(2,4); // one time initialization. 10 MHz, fractional N
			if(NASA){
				// special board configuration. Main LO on RF_OUT_B, RF_OUT_A is divided by 8
				// requires moving R33/R34 to R31/R32
				program_tuner_pll(2,6);
			} else {
				program_tuner_pll(2,4); // Program the default frequency
			}
		}
	}
}

// check TX_EN input on pin A17 (P0[16]). Control the RF_OUT_EN at the quadrature modulator.
inline void TX_EN_poll() {
	// LATENCY: 0 - 9us
	if(FIO0PIN & 0x00010000){
		PIN_RFOUT_ENB_L();	// enable the RF modulator output (active low)
		PIN_SYNTH2_ENABLE_H();	// enable the RF synthesizer output
	} else {
		PIN_RFOUT_ENB_H();	// disable the RF modulator output
		PIN_SYNTH2_ENABLE_L();	// disable the RF synthesizer output
	}
}

void DAC_init(void) {
	// initialize internal 10-bit DAC

	// 1. to enable the DAC, output must be selected to appear on the related pin p0.26
	PINSEL1 = (PINSEL1 & 0xFFCFFFFF) | 0x00200000;

	// 2. peripheral clock selection for DAC.
	// select the processor clock CCLK/4 (default)
	PCLKSEL1 = PCLKSEL1 & 0xFF3FFFFF;

	// 3. enable double buffering, no interrupts, 1us refresh
	DACCTRL = 0x00000006;
	DACCNTVAL = 30; // 1us @120MHz/4
}

void DAC(long value) {
	// output is value * Vrefp / 1024
	// 1us settling time -> 1 MHz maximum update rate
	DACR = (value << 6) & 0x0000FFC0;
}

void PWM_init(void) {
	// initialized P1[26] = PWM 1.6 as a fake 10-bit DAC (externally filtered)
	// PWM1IR interrupt register ignored (no interrupts)
	PCONP |= (1 << 6); // PWM1 power ON
	// PCLKSEL0: by default peripheral clock is CCLK/4 = 30 MHz
	PWM1TCR = 0x00000002; // Reset counter
	PWM1PR = 0;
	PWM1PC = 0;
	PINSEL3 = (PINSEL3 & 0xFFCFFFFF) | 0x00200000;
	// p1[26] = PWM1.6
	PINMODE3 = (PINMODE3 & 0xFFCFFFFF) | 0x00200000;
	// P1[26] PWM has no PU/PD
	PWM1MR0 = 1024; // controls the PWM cycle rate. Here 1024 peripheral clocks
	PWM1MR6 = 0; // initial PWM value
	// PWM Match Control Register
	// Match 0 is configured to reset the timer/counter when a match event occurs.
	PWM1MCR = 0x00000002; // reset counter when match MR0
	// PWM Control Register
	PWM1PCR = 0x00004000; // The PWM6 output enabled
	// PWM Timer Control Register
	PWM1TCR = 0x00000009; // clear counter,  enable PWM mode, enable counting
}

// control PWM1.6
void PWM(long value) {
	// output is value * 3.3V / 1024 after filtering
	PWM1MR0 = 1024; // controls the PWM cycle rate. Here 1024 peripheral clocks
	PWM1MR6 = value & 0x000003FF; // controls the PWM edge position

	// PWM Latch Enable Register
	PWM1LER = 0x00000041; // Enable PWM Match 0 and 6 Latch
}

// call once after tx is enabled
inline void setTxGain(void){
	int c;
	// tx gain set by control register. fixed
	// transmitter gain (10-bit)
	c = (int)(regv[5]) & 0x00000003;
	c <<= 8;
	c |= (int)(regv[4]) & 0x000000FF;
	DAC(c);
}




//long pinDiodeAttenuatorCompensation(long gain){
//	// attempt at avoiding instabilities
//	// function of center frequency fc (global)
//	// input range 0 - 1023
//
//	long revised_gain;
//	revised_gain = gain;
//
//	if(fc < 200000000){
//		// below 200 MHz, avoid very large attenuations (Vc < 0.55V) which are destabilizing
//		if(gain < 129)
//			revised_gain = 129;
//	} else if (fc < 500000000){
//		// below 500 MHz, avoid very large attenuations (Vc < 0.5V) which are destabilizing
//		if(gain < 108)
//			revised_gain = 108;
//	} else if (fc > 2000000000){
//		// above 2 GHz, avoid too much gain (> 2.5V) because it can create instabilities
//		if(gain > 538)
//			revised_gain = 538;
//	}
//	return revised_gain;
//}

void ADC_init(void) {
	// ADC initialization
	PCONP |= 0x00001000; // power ADC subsystem
	// PCLKSEL0: by default peripheral clock is CCLK/4 = 30 MHz
	// A/D Control Register: inputs are AD0[2] and [5], ADC clock is 10 MHz, burst=1, operational,
	AD0CR = 0x00210224;
}

void ADC_TX_POWER_DET(void) {
	// returns the last value read on TX_POWER_DET = AD0[5]
	unsigned long a;
	a = AD0DR5;
	if (a & 0x80000000) {
		// done bit high. update tx_power_det
		tx_power_det = (a >> 4) & 0x00000FFF; // 12-bit ADC measurement
	}
}

void ADC_TEMP(void) {
	// returns the last value read on TEMP = AD0[2]
	unsigned long a;
	a = AD0DR2;
	if (a & 0x80000000) {
		// done bit high. update temperature_det
		temperature_det = (a >> 4) & 0x00000FFF; // 12-bit ADC measurement
	}
}



