/*-------------------------------------------------------------
 *  Copyright (c) 2010, MSS ComBlock. All rights reserved.
 *
 * 	Filename: usb0.c
 *  Last modified: 8/23/10
 *  Authors:
 *  	Alain Zarembowitch / MSS ComBlock
 *  Version: 1.0
 *  Inheritance: n/a
 *  Description: USB configuration as per Chapter 11: LPC17xx USB device controller
 *
 *
 *  Dependencies/Pre-requisite: N/A
 -------------------------------------------------------------*/
#include "main.h"
#include "usb0.h"
#include "usb1.h"
#include "vectors.h"

// Chapter 11. Section 11. Interrupt handling
WEAK void USB_IRQHandler(void) {
	long intSt, devSt, epSt, pepNo, epNo;

	intSt = DEV_INT_STAT; // read interrupt source
	DEV_INT_CLR = intSt; // clear interrupt

	// process interrupt: only two sources enabled: DEV_STAT and EP_SLOW
	if (intSt & 0x00000008) {
		// DEV_STAT: USB Bus reset, USB suspend change or Connect change event occurs
		// Get Device Status from SIE
		devSt = getDeviceStatus();
		if (devSt & 0x00000010) {
			// bus reset
			//???
		}
		if (devSt & 0x00000002) {
			// connect change (read bit0 for connection status)
			//???
		}
		// ignore suspend change (suspend/resume automatically done by hardware)
	}

	if (intSt & 0x00000004) {
		// EP_SLOW Slow endpoints interrupt
		// check all 31 physical endpoints until all clear
		for (pepNo = 0; pepNo < 32; pepNo++) {
			epNo = pepNo >> 1;
			if (!EP_INT_STAT)
				break; // all interrupts cleared. No need to go through the complete list of possible endpoints
			if (EP_INT_STAT & (1 << pepNo)) {
				// found one endpoint interrupt
				EP_INT_CLR = (1 << pepNo); // clear interrupt for this endpoint
				// On completion of the command, the CDFULL bit is set, USBCmdData contains the status of the endpoint
				// wait for Command data register (USBCmdData) to be full (Data can be read now).
				while ((DEV_INT_STAT & 0x00000020) == 0)
					;
				epSt = CMD_DATA; // endpoint status

				// pass the event to the state machine
				if (epNo == 0)
					USB_ep0(pepNo & 0x01, epSt);
				else if (epNo == 2)
					USB_ep2(pepNo & 0x01, epSt);
				// other endpoints are currently unused
			}
		}
	}

}

// Chapter 11. Section 12: Serial interface engine command description
void sieCmd(unsigned long cmd){
	// 1-byte commmand, 1 byte val
	DEV_INT_CLR = 0x10; // Clear CCEMPTY.
	cmd = (cmd << 16) | 0x00000500; // CMD_PHASE=0x05(Command)
	CMD_CODE = cmd;
	while (!(DEV_INT_STAT & 0x10))
		; // Wait for CCEMPTY.
	DEV_INT_CLR = 0x10; // Clear CCEMPTY.
	return;
}

void sieCmdWrite(unsigned long cmd, unsigned long val){
	// 1-byte commmand, 1 byte val
	DEV_INT_CLR = 0x10; // Clear CCEMPTY.
	cmd = (cmd << 16) | 0x00000500; // CMD_PHASE=0x05(Command)
	CMD_CODE = cmd;
	while (!(DEV_INT_STAT & 0x10))
		; // Wait for CCEMPTY.
	DEV_INT_CLR = 0x10; // Clear CCEMPTY.
	val = (val << 16) | 0x00000100; 	// CMD_PHASE=0x01(Write)
	CMD_CODE = val;
	while (!(DEV_INT_STAT & 0x10))
		; // Wait for CCEMPTY.
	DEV_INT_CLR = 0x10; // Clear CCEMPTY.
	return;
}

long sieCmdRead(unsigned long cmd){
	// 1-byte commmand
	long a;
	DEV_INT_CLR = 0x30; // Clear both CCEMPTY & CDFULL
	a = (cmd << 16) | 0x00000500; // CMD_PHASE=0x05(Command)
	CMD_CODE = a;
	while (!(DEV_INT_STAT & 0x10))
		; // Wait for CCEMPTY.
	DEV_INT_CLR = 0x10; // Clear CCEMPTY interrupt bit.
	a = (cmd << 16) | 0x00000200; // CMD_PHASE=0x02(Read)
	while (!(DEV_INT_STAT & 0x20))
		; // Wait for CDFULL.
	DEV_INT_CLR = 0x20; // Clear CDFULL.
	a = CMD_DATA; // Read response
	return a;
}

void setAddress(long addr) {
	// set address and DEV_EN
	addr |= 0x00000080;	// combine addr, DEV_EN
	sieCmdWrite(0x000000D0,addr);
	return;
}

long getDeviceStatus(void) {
	return sieCmdRead(0x000000FE);
}

void selectEndpoint(long pepNo) {
	// select physical endpoint pepNo
	pepNo &= 0x0000001F;	// valid range 0 - 1F
	sieCmd(pepNo);
	return;
}
void clearBuffer(void) {
	// clear buffer (one must use selectEndpoint() first)
	sieCmd(0x000000F2);
	return;
}

void clearStallEndpoint(long pepNo){
	// clear the stall bit for the selected physical endpoint
	pepNo &= 0x0000001F;	// valid range 0 - 1F
	sieCmdWrite(pepNo, 0);
	return;
}

void setStallEndpoint(long pepNo){
	// set the stall bit for the selected physical endpoint
	pepNo &= 0x0000001F;	// valid range 0 - 1F
	sieCmdWrite(pepNo, 0x00000001);
	return;
}

void validateBuffer(void){
	// When the CPU has written data into an IN buffer, software should issue a Validate Buffer command
	sieCmd(0x000000FA);
	return;
}

// Chapter 11. Section 13. USB device controller initialization
void usbInit(void) {
	// enable the device controller clocks
	USBClkCtrl = 0x00000012;
	// poll until they are set
	while (!(USBClkSt == 0x00000012))
		;
	usbReset();
	return;
}

void usbReset(void) {
	// See Chapter 11, section 13: USB device controller initialization
	// Enable physical endpoints 0,1 (control). Other endpoints will be enabled upon request by the USB host.


	EP_INDEX = 0; // point to physical endpoint EP0 (control, out)
	MAXPACKET_SIZE = USB_MAX_PACKET0; // set max packet size
	EP_INDEX = 1; // point to EP1 (control, in)
	MAXPACKET_SIZE = USB_MAX_PACKET0;
	// poll until the endpoints are realized
	while (!(DEV_INT_STAT & 0x00000008))
		;

	// Enable endpoint interrupts (Slave mode)
	EP_INT_CLR = 0xFFFFFFFF; // Clear all endpoint interrupts
	EP_INT_EN = 0xFFFFFFFF; // Enable Slave mode for all (no DMA)
	DEV_INT_CLR = 0xFFFFFFFF; // Clear any device interrupts
	// USBEpIntPri default -> interrupts rounted to EP_SLOW
	DEV_INT_EN = 0x0000000C; // device interrupts enabled: DEV_STAT, EP_SLOW
	setAddress(0);

	return;
}

// Chapter 11. Section 14. Slave mode operation
long usbReadEp(long pepNo, unsigned char * buf, long bufsize) {
	long rxLength, i, a, regval;

	// read endpoint data
	regval= ((pepNo & 0x001E) << 1) | 0x00000001; // RD_EN and logical endpoint
	 USB_CTRL = regval;
	// wait until the PKT_RDY packet length information is ready
	while (!(RX_PLENGTH & 0x00000800))
		;
	rxLength = RX_PLENGTH & 0x000003FF; // length in bytes in bits(9:0)

	for (i = 0; i < rxLength; i += 4) {
		a = RX_DATA;
		if ((i + 4) > bufsize) // don't write past the buffer declared size
			break;
		memcpy(&buf[i], &a, 4);
	}
	USB_CTRL = regval & (~0x00000001);	// clear RD_EN bit
	// clear buffer command
	selectEndpoint(pepNo);
	clearBuffer();
	return rxLength;
}

void usbWriteEp(long pepNo, unsigned char * buf, long nBytes2write) {
	long i, a, regval;

	// write endpoint data
	regval= ((pepNo & 0x001E) << 1) | 0x00000002; // WR_EN and logical endpoint
	 USB_CTRL = regval;
	 TX_PLENGTH = nBytes2write;


	  for (i = 0; i < nBytes2write; i += 4) {
		memcpy(&a, &buf[i], 4);
 	    TX_DATA = a;
	  }
	  USB_CTRL = 0;
	  selectEndpoint(pepNo);
	  validateBuffer();
	  return;
}
