/*-------------------------------------------------------------
 *  Copyright (c) 2010-2017, MSS ComBlock. All rights reserved.
 *
 * 	Filename: utils.c
 *  Last modified: 2/5/17
 *  Authors:
 *  	Alain Zarembowitch / MSS ComBlock
 *  Version: 3
 *  Inheritance: N/A
 *  Description: Utilities for ARM Cortex-M3 LPC17xx
 *  Dependencies/Pre-requisite: N/A
 *
 *  Rev3 12/12/15 AZ
 *  Added usb connection/disconnection in main, depending on the presence of the USB VBUS
 -------------------------------------------------------------*/
#include "main.h"
#include "utils.h"
#include "uart.h"

long JP1val;	// Jumper JP1 position 0(none), 12 or 23. Call JP1_jumper() to assess

int power_check(void){
	unsigned int a;
	a = 0;
	if(FIO1PIN & 0x00000001)		// pin P1[0] PGOOD1 ARM/D_+3.3V
		a |= 0x01;
	if(FIO1PIN & 0x00000002)		// pin P1[1] PGOOD2 MOD_+4.75V
		a |= 0x02;
	if(FIO1PIN & 0x00000010)		// pin P1[4] PGOOD3 TX_SYNTH +3.3V
		a |= 0x04;
	if(FIO1PIN & 0x00000100)		// pin P1[8] PGOOD4 D_+1.8V
		a |= 0x08;
	if(FIO1PIN & 0x00000200)		// pin P1[9] PGOOD5 DACA_+3.3V
		a |= 0x010;
	if(FIO1PIN & 0x00800000)		// pin P1[23] PGOOD6 ATTN_+1.2V
		a |= 0x020;
	return a;
}

void delay_ms(unsigned long delay){
	// delay expressed in ms.
	// Step size is 10ms.
	// A key assumption is that the system timer is configured for a 10ms time interval.
	// Remember to reconfigure the system time if the CPU clock frequency is modified (see setSystemTimer)
	unsigned long startTime;
	if(delay == 0) return;
	startTime = SysTickCnt;
	delay = (delay / 10) + 1;  // normalize to 10ms increments = tick period. round up.
	while ((SysTickCnt - startTime) < delay);
	return;
}

void delay_ns(unsigned long delay){
	// delay expressed in ns ASSUMMING a 120 MHz CPU clock
	// step size is 8ns
	// Overhead when calling this function: TBD  // TODO measure overhead in situ
	unsigned long startTime;
	if(delay == 0) return;
	startTime = STCURR;
	delay >>= 3; // divide by 8(ns) = approx 1/120 MHz CPU clock period.
	// Note: STCURR is a count-DOWN counter!
	while((startTime - STCURR) < delay){;}
	return;
}

void sendLine(long destPortNo){
	// send txLine to the port identified by destPortNo:  1 = top, 2 = right, 3 = bottom, 4 = left (ComBlock rule)
	txLine.rptr = 0;	// mark message as unread
	txLine.msg_complete = 1;	// mark message as complete
	txLine.portNo = destPortNo;	// destination port
	if(destPortNo == 1){
		// to LPC17xx ARM USB (M&C pipe)
		USB_ARM_mctx();
	} else if((destPortNo == 2) || (destPortNo == 3) || (destPortNo == 4)){
		// to UARTs right/bottom/left connector
		uart_txpoll();
	}
	return;

}

void forwardLine(void){
	// forwarding destination defined by forwardMask
	// Never forward back to the sender

	// copy message to txLine
	memcpy(txLine.buf, rxLine.buf, rxLine.wptr);
	txLine.wptr = rxLine.wptr;

// Do not forward to USB, only to LVTTL serial pins
//	if((!(forwardMask & 0x01)) && (rxLine.portNo != 1)){
//		sendLine(1);
//	}
	if((!(forwardMask & 0x02)) && (rxLine.portNo != 2)){
		sendLine(2);
	}
	if((!(forwardMask & 0x04)) && (rxLine.portNo != 3)){
		sendLine(3);
	}
	if((!(forwardMask & 0x08)) && (rxLine.portNo != 4)){
		sendLine(4);
	}
	return;

}

int dispatchLine(void){
	// check what to do with the rxLine:
	// forward, process locally, pass other ComBlock response back to the PC?
	// return 1 if the message is to be processed locally, 0 otherwise


	// rx message is addressed to all (111 is the broadcast address). forward it then act upon it locally
	if (!memcmp(&rxLine.buf[1], "111",3)){
		forwardLine();
		return 1;
	}

	// rx message is a response from an adjacent ComBlock to a query from the user PC (999 is the user PC address).
	// forward to the command port.
	if (!memcmp(&rxLine.buf[1], "999",3)){
		memcpy(txLine.buf, rxLine.buf, rxLine.wptr);
		txLine.wptr = rxLine.wptr;
		sendLine(commandPortNo);
		// clear rx buffer
		rxLine.rptr = rxLine.wptr;
		return 0;	// do not process locally
	}

	// rx message is for us: address at the start of the rx message matches exactly
	if (!memcmp(&rxLine.buf[1], &address[0],3)){
		return 1;
	}

	// rx message is destined for a specific Comblock but not for us. forward it
	forwardLine();
	// clear rx buffer
	rxLine.rptr = rxLine.wptr;
	return 0;	// do not process locally

}


unsigned char hex2bin(char istring[2])
{
	unsigned char a;
	unsigned char b;
	a = istring[0];
	b = istring[1];

	if ((a >= 'A') && (a <= 'F'))
		a = a - 'A' + 10;
	else
		a = a - '0';

	if ((b >= 'A') && (b <= 'F'))
		b = b - 'A' + 10;
	else
		b = b - '0';

	a = (((a << 4) & 0xf0) | (b & 0x0f));
	return a;
}

void char2hex(unsigned char input, char * buf2){
	// converts an unsigned char (input) into a 2-character string (buf2), not terminated (no \0 \r \n)
	// equivalent to sprintf(  %2X) format.... library is either too big or cause problems.
	unsigned char a;
	a = input & 0x0F;
	if(a < 10){
		buf2[1] = '0' + a;
	} else {
		buf2[1] = 'A' + a - 10;
	}
	a = (input >> 4) & 0x0F;
	if(a < 10){
		buf2[0] = '0' + a;
	} else {
		buf2[0] = 'A' + a - 10;
	}
	return;
}

void toggle1(int nToggle){
	// toggle pin P1[1] pin75 nToggle times. Used for real-time debugging. see schematics
	int i;
	for (i=0; i<nToggle; i++){
		FIO1SET = 0x00000002;
		FIO1CLR = 0x00000002;
	}
}

void toggle2(int nToggle){
	// toggle pin P2[10] (J4 pin1) nToggle times.
	// Used for real-time debugging.
	int i;
	for (i=0; i<nToggle; i++){
		FIO2SET = 0x00000400;
		FIO2CLR = 0x00000400;
	}
}

long get_JP1val(void){
	// read the JP1 jumper setting (=user's intent)
	// One must first call assess_JP1 (which takes time) whereas get_JP1val immediately returns a value.
	// returns:
	//   0 when no jumper
	//   12 when jumper in position 1-2
	//   23 when jumper in position 2-3
	return JP1val;
}

void assess_JP1(void){
	// read the JP1 jumper setting (=user's intent).
	// and save to JP1val;
	//   0 when no jumper
	//   12 when jumper in position 1-2
	//   23 when jumper in position 2-3

	// remember P1[20] direction and functionality
	long ps, dir, a;

	// save current configuration
	ps = PINSEL3;
	dir = FIO1DIR;

	// P1[20] is now GPIO input
	FIO1DIR &= 0xFFEFFFFF;
	PINSEL3 &= 0xFFFFFCFF;

	PINMODE3 = PINMODE3 & 0xFFFFFCFF ;	// pull-up pin 2 P1[20]
	// wait a little bit (pull-up is not that strong)
	delay_ns(10000);
	a = FIO1PIN;
	if((a & 0x00100000) == 0){
		JP1val = 12;	// jumper is in position 1-2
		goto restore;
	}
	PINMODE3 = PINMODE3 | 0x00000300 ;	// pull-down pin 2 P1[20]
	// wait a little bit (pull-up is not that strong)
	delay_ns(10000);
	a = FIO1PIN;
	if(!((a & 0x00100000) == 0 )){
		JP1val = 23;	// jumper is in position 2-3
		goto restore;
	}
	JP1val = 0;	// no jumper

restore:
	// restore P1[20] direction and functionality
	PINSEL3 = ps;
	FIO1DIR = dir;
	return;

}

// check USB VBUS level
inline int USB_VBUS_check(void) {
	return FIO1PIN & 0x40000000; // pin P1[30] USB VBUS
}

/*
 *  USB Connect Function
 *   Called by the User to Connect/Disconnect USB
 *    Parameters:      con:   Connect/Disconnect
 *    Return Value:    None
 */
void USB_Connect (unsigned long con) {
  // controls a MOSFET switch to enable/disable 1.5 KOhm PU by software.
  // signal is inverted (USB_CONNECT#) (same code in COM-3011,COM-1500,etc)
  if(con){
	  // connect
	  PIN_USB_CONNECTB_L();
  } else {
	  // disconnect
	  PIN_USB_CONNECTB_H();
  }
  // original code in usbhw.c
}

// control USB_CONNECTB based on the confirmed presence or absence of VBUS voltage from the USB host
void USB_connect_ctrl(void){
	if((usb_softconnect == 0) && (USB_VBUS_check())){
		// VBUS presence detected while USB is unconnected
		delay_ms(2);	// wait to confirm VBUS presence
		if(USB_VBUS_check()){
			// VBUS presence confirmed
			usb_softconnect = 1;
			USB_Init();
		}
	} else if((usb_softconnect == 1) && (!USB_VBUS_check())){
		// first VBUS absence detected while USB is connected
		delay_ms(2);	// wait to confirm VBUS absence
		if(!USB_VBUS_check()){
			// VBUS absence confirmed
			usb_softconnect = 0;
			USB_Connect(0);
		}
	}
}

